# Copyright 2013-2014 University of Warsaw
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import os.path
import sys
import errno

def hdfs_filesystem_warning():
    return 'maybe you did not specify the right file system? '\
        'Remember that if you want to access files in the local file system, '\
        'you have to prefix the path with "{}" and '\
        'if you want to access HDFS, no prefix should be given.'\
        .format(FileSystemPathFactory.local_fs_path_prefix)

def import_hdfs_lib():
    """Imports required modules from pydoop package locally. 
    
    This way, if we don't want to use the functionality of accessing HDFS, 
    having pydoop installed in the system is not necessary.
    """
    try:
        import pydoop
    except ImportError:
        print('ERROR: It seems that you want to access HDFS. '\
              'The module "pydoop" needs to be installed in the system '\
              'in order to do it. See avroknife packages\'s README file or '\
              'package\'s description in the http://pypi.python.org/ repository '\
              '(section "Installation") for further details.\n'\
              'On the other hand, {}'\
            .format(hdfs_filesystem_warning()), file=sys.stderr)
        sys.exit(1)
    ## We're importing stuff from pydoop package only locally, 
    ## so if we don't want to use the functionality of the
    ## script of accessing HDFS, having pydoop installed in the system
    ## is not required.
    import pydoop.hdfs as hdfs
    import pydoop.hdfs.path as hdfspath
    return (hdfs, hdfspath)

class FileSystemPathFactory:
    local_fs_path_prefix = "local:"

    @staticmethod
    def create(path):
        if path.startswith(FileSystemPathFactory.local_fs_path_prefix):
            sub = path[len(FileSystemPathFactory.local_fs_path_prefix):]
            return LocalPath(sub)
        else:
            return HDFSPath(path)

class FileSystemPath:
    """File system and path abstraction"""
    
    def open(self, mode="r"):
        """Open file

        The returned file-like object must support "seek" operation, so 
        returning just a stream is not enough. 
        """
        raise NotImplementedError

    def ls(self):
        raise NotImplementedError

    def exists(self):
        raise NotImplementedError

    def is_dir(self):
        raise NotImplementedError

    def append(self, string):
        raise NotImplementedError

    def make_dirs(self):
        raise NotImplementedError

    def __lt__(self, other):
        return str(self) < str(other)

class LocalPath(FileSystemPath):
    def __init__(self, path):
        self.__path = path

    def open(self, mode="r"):
        return open(self.__path, mode=mode)
    
    def ls(self):
        return os.listdir(self.__path)

    def exists(self):
        return os.path.exists(self.__path)
    
    def is_dir(self):
        return os.path.isdir(self.__path)
    
    def append(self, string):
        return LocalPath(os.path.join(self.__path, string))

    def make_dirs(self):
        try:
            os.makedirs(self.__path)
        except OSError as ex:
            if ex.errno == errno.EEXIST and os.path.isdir(self.__path):
                pass

    def __str__(self):
        return self.__path

class HDFSPath(FileSystemPath):
    def __init__(self, path):
        (hdfs, hdfspath) = import_hdfs_lib()
        self.hdfs = hdfs
        self.hdfspath = hdfspath
        self.__path = path

    def open(self, mode="r"):
        return self.hdfs.open(self.__path, mode)
    
    def ls(self):
        files = []
        for absolute_path in self.hdfs.ls(self.__path):    
            files.append(os.path.basename(absolute_path))
        return files
    
    def exists(self):
        return self.hdfspath.exists(self.__path)
    
    def is_dir(self):
        return self.hdfspath.isdir(self.__path)
    
    def append(self, string):
        return HDFSPath("{}/{}".format(self.__path, string))

    def make_dirs(self):    
        self.hdfs.mkdir(self.__path)
            
    def __str__(self):
        return self.__path


# Copyright 2013-2014 University of Warsaw
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import uuid
import tempfile
import shutil

from avroknife.file_system import import_hdfs_lib

class FileSystem:
    """File system abstraction"""

    def create_temporary_dir(self, dir=None):
        """
        Args:
            dir: If dir is specified, the file will be created in that directory; 
                otherwise, a default directory is used.

        Returns:
            path to the directory
        """
        raise NotImplementedError

    def delete_dir(self, path):
        raise NotImplementedError

    def create_dir(self, path):
        raise NotImplementedError

    def copy_from_local_dir(self, src_dir, dst_dir):
        """Copy directory from a local file system
        
        Args:
            dst_dir: the destination directory; it mustn't already exist.
        """
        raise NotImplementedError

    def copy_to_local_dir(self, src_dir, dst_dir):
        """Copy directory to a local file system
        
        Args:
            dst_dir: the destination directory; it mustn't already exist.
        """
        raise NotImplementedError

    def join_path(self, path_elements):
        raise NotImplementedError

class LocalFS(FileSystem):
    """Local file system"""

    def create_temporary_dir(self, dir=None):
        return tempfile.mkdtemp(dir=dir)

    def delete_dir(self, path):
        shutil.rmtree(path)

    def create_dir(self, path):
        os.makedirs(path)

    def copy_from_local_dir(self, src_dir, dst_dir):
        shutil.copytree(src_dir, dst_dir) 

    def copy_to_local_dir(self, src_dir, dst_dir):
        self.copy_from_local_dir(src_dir, dst_dir)

    def join_path(self, path_elements):
        return os.path.join(*path_elements)

class HDFS(FileSystem):
    """Hadoop Distributed File System"""

    __default_tmp_dir = ['.tmp-avroknife']

    def __init__(self):
        (hdfs, hdfspath) = import_hdfs_lib()
        self.hdfs = hdfs
        self.hdfspath = hdfspath

    def create_temporary_dir(self, dir=None):
        file_exists = True
        path = None
        while file_exists:
            id_ = str(uuid.uuid4())
            if dir is not None:
                path = self.join_path([dir, id_])
            else:
                path = self.join_path(self.__default_tmp_dir + [id_])
            file_exists = self.hdfspath.exists(path)
        self.create_dir(path)
        return path

    def delete_dir(self, path):
        self.hdfs.rmr(path)

    def create_dir(self, path):
        self.hdfs.mkdir(path)

    def copy_from_local_dir(self, src_dir, dst_dir):
        self.hdfs.put(src_dir, dst_dir)

    def copy_to_local_dir(self, src_dir, dst_dir):
        self.hdfs.get(src_dir, dst_dir)

    def join_path(self, path_elements):
        return '/'.join(path_elements)



# Copyright 2013-2014 University of Warsaw
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import avro.schema
from avro.datafile import DataFileWriter
from avro.io import DatumWriter

def create(standard_out_path, nested_out_path, binary_out_path):
    """Create example Avro data stores"""

    __create_standard(standard_out_path)
    __create_nested(nested_out_path)
    __create_binary(binary_out_path)

def __create_standard(out_path):
    os.makedirs(out_path)
    schema_path = os.path.join(os.path.dirname(__file__), 'data/user.avsc')
    schema = avro.schema.parse(open(schema_path).read())
  
    with DataFileWriter(open(os.path.join(out_path, 'part-m-00000.avro'), 'w'), 
            DatumWriter(), schema) as writer:
        writer.append({'position': 0, 'name': 'Alyssa', 'favorite_number': 256})
        writer.append({'position': 1, 'name': 'Ben', 'favorite_number': 4, 'favorite_color': 'red'})
    
    with DataFileWriter(open(os.path.join(out_path, 'part-m-00001.avro'), 'w'), 
            DatumWriter(), schema) as writer:
        writer.append({'position': 2, 'name': 'Alyssa2', 'favorite_number': 512})
        writer.append({'position': 3, 'name': 'Ben2', 'favorite_number': 8, 'favorite_color': 'blue', 'secret':b'0987654321'})
        writer.append({'position': 4, 'name': 'Ben3', 'favorite_number': 2, 'favorite_color': 'green', 'secret':b'12345abcd'})
    
    with DataFileWriter(open(os.path.join(out_path, 'part-m-00002.avro'), 'w'), 
            DatumWriter(), schema) as writer:
        pass
    
    with DataFileWriter(open(os.path.join(out_path, 'part-m-00003.avro'), 'w'), 
            DatumWriter(), schema) as writer:
        writer.append({'position': 5, 'name': 'Alyssa3', 'favorite_number': 16})
        writer.append({'position': 6, 'name': 'Mallet', 'favorite_color': 'blue', 'secret': b'asdfgf'})
        writer.append({'position': 7, 'name': 'Mikel', 'favorite_color': ''})
    
def __create_nested(out_path):
    os.makedirs(out_path)
    schema_path = os.path.join(os.path.dirname(__file__), 'data/nested.avsc')
    schema = avro.schema.parse(open(schema_path).read())
    with DataFileWriter(open(os.path.join(out_path, 'part-m-00004.avro'), 'w'), 
                    DatumWriter(), schema) as writer:
            writer.append({'sup': 1, 'sub':{'level2':2}})
            writer.append({'sup': 2, 'sub':{'level2':1}})

def __create_binary(out_path):
    os.makedirs(out_path)
    schema_path = os.path.join(os.path.dirname(__file__), 'data/binary.avsc')
    schema = avro.schema.parse(open(schema_path).read())
    with DataFileWriter(open(os.path.join(out_path, 'content.avro'), 'w'), 
                    DatumWriter(), schema) as writer:
            various_stuff_data = open(os.path.join(os.path.dirname(__file__), 
                            'data/binary_stuff/various_stuff.tar.gz')).read()
            writer.append({'description': 'various stuff', 
                           'packed_files': various_stuff_data})
            greetings_data = open(os.path.join(os.path.dirname(__file__), 
                            'data/binary_stuff/greetings.tar.gz')).read()
            writer.append({'description': 'greetings', 
                           'packed_files': greetings_data})


# Copyright 2013-2014 University of Warsaw
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import json

from avro.datafile import DataFileWriter
from avro.io import DatumWriter

from avroknife.error import error
from avroknife.utils import encapsulate_strings, dict_to_json, to_byte_string

def to_json(data_store, record_selector, printer, pretty=False):
    """Converts selected records to JSON.
    
    Args:
        data_store: a DataStore object
        record_selector: a RecordSelector object that defines which records
            should be processed.
        printer: a Printer object that is used to print the JSON.
        pretty: True if the output should be a valid, pretty-printed JSON.
    """
    first_record = True
    if pretty:
        printer.print("[", end="") 
    for record in record_selector.get_records(data_store):
        try:
            if first_record:
                first_record = False
            else:
                if pretty:
                    printer.print(",")
                else:
                    printer.print("")
            content = encapsulate_strings(record.content)
            printer.print(dict_to_json(content, pretty), end="")
        except Exception:
            error("while processing record with index {}".format(record.index))
            raise
    if pretty:
        printer.print("]")
    else:
        ## The case when no records were present
        if not first_record:
            printer.print("")

def extract(data_store, record_selector, field_name, output_dir_path=None):
    """Extract specified field from selected records

    Args:
        data_store: a DataStore object
        record_selector: a RecordSelector object. It defines which records
            should be processed.
        field_name: name of the field to be extracted. Can contain nested names,
            e.g. 'field1.field2' .
        output_dir_path: a FileSystemPath object. This is where the extracted
            fields will be saved. If this is not given, the extracted fields
            are printed to stdout.
    """
    if output_dir_path is not None:
        output_dir_path.make_dirs()
    for record in record_selector.get_records(data_store):
        datum = record.content
        if datum: #check if there is any data to be written
            parts = field_name.split('.')
            for part in parts:
                datum = datum[part]
            if output_dir_path is not None:
                try:
                    with output_dir_path.append(str(record.index)).open('w') \
                            as output:
                        output.write(to_byte_string(datum))
                except Exception:
                    error("while processing record with index {}"\
                        .format(record.index))
                    raise
            else:
                print(to_byte_string(datum))

def copy(data_store, record_selector, output_dir_path):
    """Dump selected records from the data store to another Avro file
    
    Args:
        data_store: a DataStore object
        record_selector: a RecordSelector object. It defines which records
            should be processed.
        output_dir_path: a FileSystemPath object. This is where the dump will
            be saved.
    """
    output_dir_path.make_dirs()
    with output_dir_path.append("content.avro").open("w") as output:
        with DataFileWriter(output, DatumWriter(), data_store.get_schema()) \
                as writer:
            records = record_selector.get_records(data_store)
            for record in records:
                writer.append(record.content)

def get_schema(data_store):
    """Get data store schema
    
    Args:
        data_store: a DataStore object
    Returns:
        JSON schema
    """
    return dict_to_json(json.loads(str(data_store.get_schema())), True)

def count(data_store, record_selector):
    """Get the number of selected records in the data store
    
    Args:
        data_store: a DataStore object
        record_selector: a RecordSelector object. It defines which records
            should be processed.
    Returns:
        number
    """
    n = 0
    for _ in record_selector.get_records(data_store):
        n = n+1
    return n


# Copyright 2013-2014 University of Warsaw
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

class Printer:
    """Output printing abstraction"""
    def print(self, text, end="\n"):
        raise NotImplementedError 

class StdoutPrinter(Printer):
    """Prints to stdout"""
    def print(self, text, end="\n"):
        print(text, end=end)

class FilePrinter(Printer):
    """Prints to file"""
    def __init__(self, fs_path):
        self.__fs_path = fs_path
        self.__f = None
        self.__already_open = False

    def print(self, text, end="\n"):
        if not self.__already_open:
            self.__f = self.__fs_path.open("w")
            self.__already_open = True
        print(text, file=self.__f, end=end)